package com.sun.org.apache.bcel.internal.classfile;

import java.io.ByteArrayInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.DataInput;

public final class Signature extends Attribute
{
    private int signature_index;
    
    public Signature(Signature c) {
        this(c.getNameIndex(), c.getLength(), c.getSignatureIndex(), c.getConstantPool());
    }
    
    Signature(int name_index, int length, DataInput input, ConstantPool constant_pool) throws IOException {
        this(name_index, length, input.readUnsignedShort(), constant_pool);
    }
    
    public Signature(int name_index, int length, int signature_index, ConstantPool constant_pool) {
        super((byte)10, name_index, length, constant_pool);
        this.signature_index = signature_index;
    }
    
    @Override
    public void accept(Visitor v) {
        v.visitSignature(this);
    }
    
    @Override
    public final void dump(DataOutputStream file) throws IOException {
        super.dump(file);
        file.writeShort(this.signature_index);
    }
    
    public final int getSignatureIndex() {
        return this.signature_index;
    }
    
    public final void setSignatureIndex(int signature_index) {
        this.signature_index = signature_index;
    }
    
    public final String getSignature() {
        ConstantUtf8 c = (ConstantUtf8)super.getConstantPool().getConstant(this.signature_index, (byte)1);
        return c.getBytes();
    }
    
    private static boolean identStart(int ch) {
        return ch == 84 || ch == 76;
    }
    
    private static void matchIdent(Signature.MyByteArrayInputStream in, StringBuilder buf) {
        int ch;
        if ((ch = in.read()) == -1) {
            throw new RuntimeException("Illegal signature: " + in.getData() + " no ident, reaching EOF");
        }
        if (!identStart(ch)) {
            StringBuilder buf2 = new StringBuilder();
            int count;
            for (count = 1; Character.isJavaIdentifierPart((char)ch); ++count, ch = in.read()) {
                buf2.append((char)ch);
            }
            if (ch == 58) {
                in.skip((long)"Ljava/lang/Object".length());
                buf.append(buf2);
                ch = in.read();
                in.unread();
            }
            else {
                for (int i = 0; i < count; ++i) {
                    in.unread();
                }
            }
            return;
        }
        else {
            StringBuilder buf2 = new StringBuilder();
            ch = in.read();
            do {
                buf2.append((char)ch);
                ch = in.read();
            } while (ch != -1 && (Character.isJavaIdentifierPart((char)ch) || ch == 47));
            buf.append(buf2.toString().replace('/', '.'));
            if (ch != -1) {
                in.unread();
            }
            return;
        }
    }
    
    private static void matchGJIdent(Signature.MyByteArrayInputStream in, StringBuilder buf) {
        matchIdent(in, buf);
        int ch = in.read();
        if (ch == 60 || ch == 40) {
            buf.append((char)ch);
            matchGJIdent(in, buf);
            while ((ch = in.read()) != 62 && ch != 41) {
                if (ch == -1) {
                    throw new RuntimeException("Illegal signature: " + in.getData() + " reaching EOF");
                }
                buf.append(", ");
                in.unread();
                matchGJIdent(in, buf);
            }
            buf.append((char)ch);
        }
        else {
            in.unread();
        }
        ch = in.read();
        if (identStart(ch)) {
            in.unread();
            matchGJIdent(in, buf);
        }
        else {
            if (ch == 41) {
                in.unread();
                return;
            }
            if (ch != 59) {
                throw new RuntimeException("Illegal signature: " + in.getData() + " read " + (char)ch);
            }
        }
    }
    
    public static String translate(String s) {
        StringBuilder buf = new StringBuilder();
        matchGJIdent(new MyByteArrayInputStream(s), buf);
        return buf.toString();
    }
    
    public static boolean isFormalParameterList(String s) {
        return s.startsWith("<") && s.indexOf(58) > 0;
    }
    
    public static boolean isActualParameterList(String s) {
        return s.startsWith("L") && s.endsWith(">;");
    }
    
    @Override
    public final String toString() {
        String s = this.getSignature();
        return "Signature: " + s;
    }
    
    @Override
    public Attribute copy(ConstantPool _constant_pool) {
        return (Attribute)this.clone();
    }
    
    private static final class MyByteArrayInputStream extends ByteArrayInputStream
    {
        MyByteArrayInputStream(String data) {
            super(data.getBytes());
        }
        
        final String getData() {
            return new String(this.buf);
        }
        
        final void unread() {
            if (this.pos > 0) {
                --this.pos;
            }
        }
    }
}
